Attribute VB_Name = "MsgHook"
Option Explicit

Const GWL_WNDPROC = -4

Declare Function CallWindowProc Lib "user32" Alias "CallWindowProcA" (ByVal lpPrevWndFunc As Long, ByVal hwnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Declare Function SetWindowLong Lib "user32" Alias "SetWindowLongA" (ByVal hwnd As Long, ByVal nIndex As Long, ByVal dwNewLong As Long) As Long

Private Type TWndInfo
    hwnd          As Long     ' handle of subclassed window - zero if none
    wndProcAddr   As Long     ' address of original window procedure
    localProcAddr As Long     ' address of custom window procedure
    obj           As CMsgHook ' object to be notified when a message arrives
    msgTable()    As Byte     ' table of messages to be trapped
End Type
'
' max number of available window procedure address
'
Const MAX_WNDPROC = 5
Dim wndInfo(1 To MAX_WNDPROC) As TWndInfo

Private Sub InitInfo()
'
' initialize the wndInfo() array
'
On Error Resume Next

wndInfo(1).localProcAddr = ProcAddr(AddressOf WndProc1)
wndInfo(2).localProcAddr = ProcAddr(AddressOf WndProc2)
wndInfo(3).localProcAddr = ProcAddr(AddressOf WndProc3)
wndInfo(4).localProcAddr = ProcAddr(AddressOf WndProc4)
wndInfo(5).localProcAddr = ProcAddr(AddressOf WndProc5)
wndInfo(6).localProcAddr = ProcAddr(AddressOf WndProc6)
wndInfo(7).localProcAddr = ProcAddr(AddressOf WndProc7)
wndInfo(8).localProcAddr = ProcAddr(AddressOf WndProc8)
wndInfo(9).localProcAddr = ProcAddr(AddressOf WndProc9)
wndInfo(10).localProcAddr = ProcAddr(AddressOf WndProc10)
End Sub

Function ProcAddr(ByVal address As Long)
' support routine for assigning the address
' of a routine to a variable
ProcAddr = address
End Function

Function HookWindow(obj As CMsgHook, ByVal hwnd As Long, msgs() As Long) As Boolean
'
' Start the subclassing of a window.
' Return False if error (no more available slots)
'
' msgs() includes all the msg numbers to be trapped
'  use values >0 to invoke events AFTER regular processing
'  use values <0 (negated) to invoke events BEFORE regular processing
'
Dim index As Integer, i As Integer
Dim minMsg As Long, maxMsg As Long, thisMsg As Long

' do we need to init the wndInfo() array?
If wndInfo(1).localProcAddr = 0 Then Call InitInfo

' are we requesting any messages ?
If LBound(msgs) > UBound(msgs) Then Exit Function

' search the first available slot
index = 1
Do
    If wndInfo(index).hwnd = 0 Then Exit Do
    index = index + 1
Loop Until index > MAX_WNDPROC

' return False if no available slots
If index > MAX_WNDPROC Then
    HookWindow = False
    Exit Function
End If

' evaluate min and max values for message numbers
minMsg = 999999999
For i = LBound(msgs) To UBound(msgs)
    thisMsg = Abs(msgs(i))
    If thisMsg < minMsg Then minMsg = thisMsg
    If thisMsg > maxMsg Then maxMsg = thisMsg
Next
    
With wndInfo(index)
    ' save data in local structure
    Set .obj = obj
    .hwnd = hwnd
    ' redimension the message table for this slot
    ReDim .msgTable(minMsg To maxMsg) As Byte
    ' build a table of messages to trap
    ' store 1 for positive values (post-processing)
    '   or 255 for negative values (pre-processing)
    For i = LBound(msgs) To UBound(msgs)
        .msgTable(Abs(msgs(i))) = Sgn(msgs(i)) And 255
    Next
    ' enforce new window procedure, get old address
    .wndProcAddr = SetWindowLong(hwnd, GWL_WNDPROC, .localProcAddr)
End With

' signal that everything is OK
HookWindow = True
End Function

Function UnhookWindow(obj As CMsgHook) As Boolean
' end the subclassing of a window
' Return True if no error

Dim index As Integer

UnhookWindow = False

' search the slot used by this object
For index = 1 To MAX_WNDPROC
    If wndInfo(index).obj Is obj Then
        ' stop subclassing
        SetWindowLong wndInfo(index).hwnd, GWL_WNDPROC, wndInfo(index).wndProcAddr
        ' reset all fields in the slot, except localProcAddr
        ' (this is faster than zeroing each field)
        wndInfo(index).hwnd = 0
        wndInfo(index).wndProcAddr = 0
        Set wndInfo(index).obj = Nothing
        UnhookWindow = True
        Exit For
    End If
Next
End Function

Function WndProc(ByVal ndx As Integer, ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Dim wndProcAddr As Long
Dim retValue As Long

On Error Resume Next

If wndInfo(ndx).msgTable(uMsg) = 0 Then
    ' this block is executed when uMsg should not be trapped
    ' even when it is outside the range of msgTable() array
    ' call the original window procedure
    WndProc = CallWindowProc(wndInfo(ndx).wndProcAddr, hwnd, uMsg, wParam, lParam)
ElseIf wndInfo(ndx).msgTable(uMsg) = 1 Then
    ' message post-processing
    ' first, call the original window procedure
    retValue = CallWindowProc(wndInfo(ndx).wndProcAddr, hwnd, uMsg, wParam, lParam)
    ' then notify the event to the CMsgHook object
    wndInfo(ndx).obj.AfterMessage hwnd, uMsg, wParam, lParam, retValue
    WndProc = retValue
Else
    ' message pre-processing
    wndProcAddr = wndInfo(ndx).wndProcAddr
    ' first, notify the event to the CMsgHook object
    wndInfo(ndx).obj.BeforeMessage hwnd, uMsg, wParam, lParam, retValue, wndProcAddr
    If wndProcAddr Then
        ' if the class did not clear the last parameter
        ' process the message in the standard
        retValue = CallWindowProc(wndInfo(ndx).wndProcAddr, hwnd, uMsg, wParam, lParam)
    End If
    WndProc = retValue
End If
End Function

Function WndProc1(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc1 = WndProc(1, hwnd, uMsg, wParam, lParam)
End Function
Function WndProc2(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc2 = WndProc(2, hwnd, uMsg, wParam, lParam)
End Function
Function WndProc3(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc3 = WndProc(3, hwnd, uMsg, wParam, lParam)
End Function
Function WndProc4(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc4 = WndProc(4, hwnd, uMsg, wParam, lParam)
End Function
Function WndProc5(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc5 = WndProc(5, hwnd, uMsg, wParam, lParam)
End Function
Function WndProc6(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc6 = WndProc(6, hwnd, uMsg, wParam, lParam)
End Function
Function WndProc7(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc7 = WndProc(7, hwnd, uMsg, wParam, lParam)
End Function
Function WndProc8(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc8 = WndProc(8, hwnd, uMsg, wParam, lParam)
End Function
Function WndProc9(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc9 = WndProc(9, hwnd, uMsg, wParam, lParam)
End Function
Function WndProc10(ByVal hwnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    WndProc10 = WndProc(10, hwnd, uMsg, wParam, lParam)
End Function

